from tkinter import *
from tkinter import messagebox
from tkinter import filedialog
import string
MAXBLOCK = 65536
SAFEBLOCK = 1024*1024
#■　2次元配列の宣言
def array2(N1,N2): return [[0 for j in range(N2)]for k in range(N1)]
#■　リトル・エンディアンの処理
def little16(c):   return c[0] + (c[1]<<8)
def little32(c):   return c[0] | (c[1]<<8) | (c[2]<<16) | (c[3]<<24)
#■　ビッグ・エンディアンの処理
#def big16(X):
#    c[0]=(X>>8) & 0xFF
#    c[1]=0xFF
#    return c
#def big32(X):
#    c[0]=(X>>24) & 0xFF
#    c[1]=(X>>16) & 0xFF
#    c[2]=(X>>8) & 0xFF
#    c[3]=X & 0xFF
#    return c
#■　メモリ・エラーを起こさない読み込み
def safe_read(fp, size):
    if size <= 0: return ""
    if size <= SAFEBLOCK: return fp.read(size) #SAFEBLOCKより小さければ一度に入力
    data = []
    while size > 0: #SAFEBLOCKより大きいとき分割入力
        block = fp.read(min(size, SAFEBLOCK))
        if not block:
            break
        for i in range(len(block)): # 2.7ではstringとして扱っていたので
            data.append(block[i])   # data.append(block)だけで良かった。
        size = size - len(block)
    return data
#■ byte表現をPython色表現に変換
def RGB(R,G,B): return "#%02X%02X%02X" % (R & 0xFF, G & 0xFF, B & 0xFF)
#■ Python色表現からR・G・B要素に分解
HexaValue=[ord('0'), ord('9'), ord('A'), ord('F'),ord('a'),ord('f')]
def HValue(X):
    D=ord(X)
    if D>=HexaValue[0] and D<=HexaValue[1]:
        return D-HexaValue[0]
    elif D>=HexaValue[2] and D<=HexaValue[3]:
        return D-HexaValue[2]+10
    elif D>=HexaValue[4] and D<=HexaValue[5]:
        return D-HexaValue[4]+10
    else:
        return 0
def RED(X)  : return HValue(X[1])*16+HValue(X[2])
def GREEN(X): return HValue(X[3])*16+HValue(X[4])
def BLUE(X) : return HValue(X[5])*16+HValue(X[6])
#■ ＤＩＢファイルヘッダ
class DIBfileHeader:
    def __init__(self):
        self.bfType=0x4D42
    #■ヘッダ読み込み
    def read(self, fp):
        fp.seek(0);
        s=fp.read(2); self.Type=little16(s)
        s=fp.read(4); self.Size=little32(s)
        s=fp.read(4);
        s=fp.read(4); self.OffsetBits = little32(s)
    #■ヘッダ表示
    def print(self):
        print("            ファイル種類： ",hex(self.Type),"(",self.Type,")")
        print("    ファイル全体のサイズ： ",hex(self.Size),"(",self.Size,")")
        print("ピクセルビットオフセット： ",hex(self.OffsetBits),"(",self.OffsetBits,")")

#■ Bitmap情報
class BitmapInfo:
    def __init__(self):
        self.bcSize=0
    #s■ビットマップ情報読み込み
    def read(self,fp):
        fp.seek(14)      # 14byte 目に移動
        s=fp.read(4); self.Size = little32(s)
        self.RGBsize=4
        if self.Size==12: # IBM BMP(BitmapInfoのサイズが12のときIBM形式)
            self.RGBsize=3
            s=fp.read(2); self.width          =  little16(s)
            s=fp.read(2); self.height         = little16(s)
            s=fp.read(2); self.Planes         = 1
            s=fp.read(2); self.BitCount       = little16(s)
        else:            # Microaoft Bmp
            s=fp.read(4); self.width          = little32(s)
            s=fp.read(4); self.height         = little32(s)
            s=fp.read(2); self.Planes         = 1
            s=fp.read(2); self.BitCount       = little16(s)
            s=fp.read(4); self.Compression    = little32(s)
            s=fp.read(4); self.SizeImage      = little32(s)
            s=fp.read(4); self.XPperMeter     = little32(s)
            s=fp.read(4); self.YPperMeter     = little32(s)
            s=fp.read(4); self.ColorUsed      = little32(s)
            s=fp.read(4); self.ColorImportant = little32(s)
    #■Bitmap情報の表示
    def print(self):
        print("                    Size : ", hex(self.Size          ),"(",self.Size,")")
        print("                   width : ", hex(self.width         ),"(",self.width,")")
        print("                  height : ", hex(self.height        ),"(",self.height,")")
        print("               Bit Count : ", hex(self.BitCount      ),"(",self.BitCount,")")
        print("             Compression : ", hex(self.Compression   ),"(",self.Compression,")")
        print("             Image Bytes : ", hex(self.SizeImage     ),"(",self.SizeImage,")")
        print("           X pixel/Meter : ", hex(self.XPperMeter    ),"(",self.XPperMeter,")")
        print("           Y pixel/Meter : ", hex(self.YPperMeter    ),"(",self.YPperMeter,")")
        print("   Number of Color Table : ", hex(self.ColorUsed     ),"(",self.ColorUsed,")")
        print("         Important Color : ", hex(self.ColorImportant),"(",self.ColorImportant,")")

#■ Bitmapクラス(カラーマップとビットマップの元の情報を含む）
class Bitmap:
    def __init__(self, header,info):
        self.header=header
        self.info=info
    #■カラーマップとビットマップの元の情報を読み込む
    def read(self,fp):
        if self.info.BitCount>=24:
            self.ColormapLen=0
            self.Colormap=[]
        else:
            self.ColorBufLen=self.info.RGBsize*(1<<self.info.BitCount)
            self.ColorBuf=fp.read(self.ColorBufLen)
            self.createColorMap()
        self.Len=self.header.Size-self.header.OffsetBits
        self.Buffer=safe_read(fp,self.Len)
    #■ファイル内のカラーマップをPythonのカラー形式に変換する
    def createColorMap(self):
        RGBsize=self.info.RGBsize
        N=1<<self.info.BitCount
        self.ColorMapLen=N
        self.ColorMap=[0 for j in range(N)]
        k=0
        pClr=self.ColorBuf
        for i in range(N):
            self.ColorMap[i]=RGB(pClr[k+2],pClr[k+1],pClr[k])
            k+=RGBsize
            
#■ Canvas表示クラス
class CBitmap:
    def __init__(self, canvas, width,height,BackgroundColor='white'):
        self.Canvas=canvas
        self.width=width
        self.height=height
        #キャンバス内を擬似的なピクセル(1×1の正方形)で埋め尽くす
        self.clearAll()
        self.ID=array2(width,height)
        self.BackgroundColor=BackgroundColor
        for Y in range(height):
            for X in range(width):
                self.ID[X][Y]=canvas.create_rectangle(X,Y,X,Y,fill=BackgroundColor, outline='')
    #■キャンバス内itemをすべてクリア
    def clearAll(self): self.Canvas.delete(ALL)
    #■24ビットＢＭＰの表示
    def draw24Bitmap(self,bitmap,w,h,Y,dY,yLoop,bMap):
        NCount=3 #24bitsのとき3バイト単位，32ビットのとき4バイト単位
        if bitmap.info.BitCount>24:  NCount=4
        k=0
        for j in range(yLoop):
            if (Y>=0) and (Y<=self.width):
                k=int((k+3)/4)*4  # 4バイト境界に合わせる(32ビットのとき変化なし)
                for X in range(w):
                    if X<=self.width:
                        CL=RGB(bMap[k+2],bMap[k+1],bMap[k])
                        self.Canvas.itemconfig(self.ID[X][Y], fill=CL)
                    k+=NCount
                    X+=1
            Y+=dY
        
    #■8ビットＢＭＰの表示
    def draw8Bitmap(self,bitmap,w,h,Y,dY,yLoop,bMap):
        NCount=1;  k=0
        for j in range(yLoop):
            if (Y>=0) and (Y<=self.width):
                k=int((k+3)/4)*4 # 4バイト境界に合わせる(32bits / 8bits=4)
                for X in range(w):
                    if X<=self.width:
                        CL=bitmap.ColorMap[bMap[k]]
                        self.Canvas.itemconfig(self.ID[X][Y], fill=CL)
                    k+=NCount
                    X+=1
            Y+=dY
     
      
    #■4ビットＢＭＰの表示
    def draw4Bitmap(self,bitmap, w,h,Y,dY,yLoop,bMap):
        NCount=1; k=0
        for j in range(yLoop):
            if (Y>=0) and (Y<=self.width):
                k=int((k+7)/8)*8 # 4バイト境界に合わせる(32bits / 4bits = 8) 
                for X in range(w):
                    if X<=self.width:
                        kk=int(k/2)  
                        if (k & 1): # 偶数番目のとき上位4bits
                            CL=bitmap.ColorMap[bMap[kk] & 0xF]
                        else:       # 奇数番目のとき下位4bits
                            CL=bitmap.ColorMap[bMap[kk] >> 4]
                        self.Canvas.itemconfig(self.ID[X][Y], fill=CL)
                    k+=NCount
                    X+=1
            Y+=dY
    #■1ビットＢＭＰの表示
    def draw1Bitmap(self,bitmap,w,h,Y,dY,yLoop,bMap):
        NCount=1; k=0
        for j in range(yLoop):
            if (Y>=0) and (Y<=self.width):
                k=int((k+31)/32)*32  # 32bits境界に合わせる
                for X in range(w):
                    if X<=self.width:
                        kk=int(k/8)  # kk: バイト単位の位置
                        imap = (bMap[kk]>>(7 - (k %8))) & 1 # 0または1
                        CL=bitmap.ColorMap[imap]
                        self.Canvas.itemconfig(self.ID[X][Y], fill=CL)
                    k+=NCount
                    X+=1
            Y+=dY
    
    #■ＢＭＰの表示
    def drawBitmap(self, bitmap):
        w=bitmap.info.width
        h=bitmap.info.height
        if h<0: # 負のとき画面上位から下方向へ
            Y=0
            dY=1
            yLoop=-h
        else: # 正のとき画面下位から上方向へ
            Y=h-1
            dY=-1
            yLoop=h
        bMap=bitmap.Buffer
        if bitmap.info.BitCount>=24:
            self.draw24Bitmap(bitmap,w,h,Y,dY,yLoop,bMap)
        
        elif bitmap.info.BitCount==8:
            self.draw8Bitmap(bitmap, w,h,Y,dY,yLoop,bMap)
        
        elif bitmap.info.BitCount==4:
            self.draw4Bitmap(bitmap, w,h,Y,dY,yLoop,bMap)
        
        elif bitmap.info.BitCount==1:
            self.draw1Bitmap(bitmap, w,h,Y,dY,yLoop,bMap)
        
        else:
            if bitmap.info.BitCount<24:
                dsplayColorMap(self.canvas,bitmap)
            messagebox.showinfo("BmpDisplay", str(bitmap.info.BitCount)+"ビットの画像は今だサポートしていません。")
    #■ピクセル取り出し(返却値はPythonのカラー形式（string型))
    def GetPixel(self, X,Y):
        if X<0 or X>=self.width or Y<0 or Y>=self.height: return '#000000'
        return self.Canvas.itemcget(self.ID[X][Y], option='fill')
    def GetPixelList(self, X,Y):
        if X<0 or X>=self.width or Y<0 or Y>=self.height: return [0,0,0]
        CL=self.Canvas.itemcget(self.ID[X][Y], option='fill')
        return [RED(CL),GREEN(CL),BLUE(CL)]
    def SetPixel(self,X,Y, CL):
        if X<0 or X>=self.width or Y<0 or Y>=self.height: return
        self.Canvas.itemconfig(self.ID[X][Y], fill=CL)
    def SetPixelList(self,X,Y, CLL):
        if X<0 or X>=self.width or Y<0 or Y>=self.height:return
        self.Canvas.itemconfig(self.ID[X][Y], fill=RGB(CLL[0],CLL[1],CLL[2]))
       
#■ カラーマップ表示(Python用形式に変換したので，前バージョンを変更)    
def displayColormap(canvas,bitmap):
    k=0; Y=-20
    for i in range(bitmap.ColorMapLen):
        if i% 16==0: Y+=30; X=10
        CL=bitmap.ColorMap[i]
        canvas.create_rectangle(X,Y,X+20,Y+20, fill=CL,outline='black')
        X+=30
def drawBitmap(canvas, bitmap):
    cbit=CBitmap(canvas, bitmap.info.width,bitmap.info.height)
    cbit.drawBitmap(bitmap)
    return cbit
def readBitmap(filename):
    try:
        fp=open(filename,'rb')
        fp.seek(0)
        #ファイル・ヘッダ読込み
        header=DIBfileHeader()
        header.read(fp)
        header.print()
        if header.Type!=0x4D42:
            messagebox.showerror("ファイル種類", "BMPファイルではありません。")
            fp.close()
            return []
        else:
            #ファイル・ヘッダ読込み
            info=BitmapInfo()
            info.read(fp)
            info.print()
            root.title("1画素当たりのビット数 = "+str(info.BitCount))
            #カラーテーブルおよびビットマップ読み込み
            bitmap=Bitmap(header,info)
            bitmap.read(fp)
            fp.close()
            return bitmap
    except IOError:
        raise IOError("Unspported BMP header type)" )
        return []
#■ メイン処理
root=Tk()
FT=[('BMPファイルlファイル','*.txt;*.*')]
DR='c:\\'
filename=filedialog.askopenfilename(filetypes=FT)
if filename!="":
    bitmap=readBitmap(filename)
    if bitmap!=[]:
        #■ビットマップ表示
        #　(以下3行のコメントはカラーマップ表示。画像はフルカラー以外を指定する)
        #canvas=Canvas(root,width=500,height=500)
        #canvas.pack()
        #displayColormap(canvas,bitmap)
        #---カラーマップ表示するときはroot.update()の直前までをコメントにする---
        canvas=Canvas(root,width=bitmap.info.width,height=bitmap.info.height)
        canvas.pack()
        cbit=drawBitmap(canvas, bitmap)
        root.update()
        #■ GetPixelとSetPixelのテスト
        for x in range(100):
             for y in range(100):
                 cbit.SetPixel(x,y,cbit.GetPixel(100+x,100+y))
        #----------------------------------------------------------------------
        root.update()
        root.mainloop()
